
#define IDENTIFICATION "2002.08.22 CDP 2.31 William W. Patterson <ailanto@komputado.com>"

#define DEBUGGING

/******************************************************************************

Copyright (C) 2002 William W. Patterson

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Purpose:

   My kind of cd. See help(). Or "cdp -h".

Usage:

   "cdp -p path" to push current directory onto stack while switching to new one

   "cdp" to switch back to previous directory

   And other special pre/defined switches. See help().

   Enter "cdp -h" for runtime help.

Notes:

   Inspired and based upon pushd and popd, written by Robert Lake.

   The public domain getopt() is from AT&T, lightly hacked.

   Written originally for Pacific C...

      To compile with Pacific C: pacc cdp.c

   Enhanced for MSC6.0, in which there are some slicker ways of doing
   some of these things, and the executable doesn't hang a dosbox when
   run after Vim (I haven't figured out why this happens, but there
   are other programs that exhibit the same behavior, so it's not just
   me! Must be some kind of Microsoftian evil, because my only Microsoft
   compiler is also my only compiler that doesn't make this happen!),
   but I wanted to make the translation as painless as possible because
   I really don't want to spend too much more time on this problem
   today (1998.10.23), so if it's ugly, that's why. If _MSC_VER is not
   predefined, we're assuming Pacific C. Anyway...

      To compile with MSC6.0: cl cpd.c

   I also made some things more verbose and/or roundabout than they
   needed to be; an attempt to make things more clear.

   Enhanced for HPUX 10.20, should work for other unices with little or no
   further modification.

      To compile with HP C: cc -Ae -o cdp cdp.c

   The unix version creates a temporary script which must be executed
   to perform the directory switch. To do this, I create a ksh function
   for cdp...

      function cdp
         {
         if ~/bin/cdp $*
         then
            . ~/.cdpscript
         fi
         }

   POSIX shell prefers something like this (to handle the case of
   no argument, since it just bombs if you reference an environment
   variable that hasn't been set)...

      function cdp
         {
         if ~/bin/cdp ${*:-""}
         then
            . ~/.cdpscript
         fi
         }


Authors:

   Robert Lake, rlake@cs.mun.ca, http://www.ucs.mun.ca/~z63rrl/
   William W. Patterson, ailanto@komputado.com, http://www.komputado.com

History:

   1998.10.21  1.00  WWP   Hacked from Robert Lake's pushd/popd programs.
   1998.10.21  1.01  WWP   Made help() a bit more helpful.
   1998.10.21  1.02  WWP   Use DEBUGGING macros for debug messaging.
   1998.10.22  1.10  WWP   Cut! Print! It's not much bigger with the debug
                           messaging, so I'll just leave it compiled in.
   1998.10.22  1.11  WWP   If destination is drive only, be sure it has a
                           trailing backslash.
   1998.10.22  1.12  WWP   Added special personal switches.
   1998.10.23  1.13  WWP   Modified for MSC 6.00 - for some unknown reason,
                           proggies compiled with Micro C or Pacific C hang
                           a DOSBOX at startup if VIM has run beforehand.
   1998.10.23  1.14  WWP   Oops. Swat that bug!
   1999.02.03  1.18  WWP   Changed -t destination.
   1999.02.03  1.19  WWP   Added -9 switch.
   1999.02.03  1.20  WWP   Cut! Print!
   1999.07.09  1.30  WWP   Decided I'd like it better if pushing were *not*
                           the default. Instead of -9 to not push, use -p
                           to push.
   1999.07.09  1.31  WWP   Use slicker DEBUGNOTICE method.
   1999.07.09  1.32  WWP   Eliminated some seldom-used switches.
   1999.07.09  1.33  WWP   And added some others!
   1999.07.09  1.34  WWP   Added CDPI environment variable override.
   1999.07.09  1.40  WWP   Implemented option table.
   1999.07.10  1.41  WWP   Added -l Patterson Default.
   1999.07.11  1.42  WWP   Added -q Patterson Default.
   1999.07.13  1.43  WWP   Added drive to default paths, where appropriate.
   1999.09.23  1.44  WWP   Since we always exit() after help()ing, moved
                           exit() into help().
   1999.09.23  1.50  WWP   Allow relative offset to a switch. That is, we
                           can use a switchpath *and* an argument, which
                           must be relative to that switchpath. Make sense?
   1999.09.30  1.51  WWP   Cleaned up for release.
   1999.09.30  1.52  WWP   Use -D for debug, upcase okay since seldom used.
   1999.09.30  2.00  WWP   Cut! Print!
   1999.10.12  2.10  WWP   Enhanced for HPUX.
   2000.09.24  2.20  WWP   Some ccs didn't like new-style function headers.
   2000.09.24  2.21  WWP   Some ccs didn't like long multi-line strings.
   2000.09.24  2.22  WWP   Removed some HPUX directories, added others.
   2000.09.24  2.23  WWP   Added some more hejmaj directories.
   2001.07.03  2.24  WWP   Changed some default directory switches.
   2001.07.03  2.25  WWP   Changed Neosplice references to Kafejo.
   2001.08.02  2.26  WWP   Xoom/NBCi is no more.
   2001.12.12  2.27  WWP   Reorganized websites directories.
   2002.08.22  2.28  WWP   Added comment POSIX function comment, in both
                           notes and help().
   2002.08.22  2.29  WWP   Switched eddress from kafejo to komputado.
   2002.08.22  2.30  WWP   Made directories more appropriate for Goodyear.
   2002.08.22  2.31  WWP   Lone ? gives short help.

******************************************************************************/

#define UNKNOWN
#define UNIXCRLF

#include <stdio.h>
#include <stdlib.h>

#ifdef hpux
#ifndef UNIX
#define UNIX
#endif
#endif

#ifdef _HPUX_SOURCE
#ifndef UNIX
#define UNIX
#endif
#endif

#include <string.h>
#include <ctype.h>

#ifdef _MSC_VER
#undef UNKNOWN
#include <conio.h>
#include <direct.h>
#endif

#ifdef HI_TECH_C
#undef UNKNOWN
#include <conio.h>
#include <unixio.h>
#include <stat.h>
#include <sys.h>
#endif

#ifdef UNIX
#undef UNKNOWN
#include <malloc.h>
#include <unistd.h>
#undef UNIXCRLF
#define UNIXCRLF puts("")
#endif

#define TRUE 1
#define FALSE 0

#define SEEK_SET 0
#define SEEK_CUR 1
#define SEEK_END 2

#ifdef DEBUGGING
#define DEBUGNOTICE
#define DEBUGMESSAGE(a) if (debugging) printf a
#else
#define DEBUGNOTICE puts("cdp was compiled without debug messaging")
#define DEBUGMESSAGE(a)
#endif

/* We should probably use PATH_MAX or _POSIX_PATH_MAX or some such for */
/* unices, but this should suffice...                                  */

#ifdef _MSC_VER
#define MAXPATHLENGTH _MAX_PATH
#else
#define MAXPATHLENGTH 260 /* Including drive and terminator, if dos */
#endif

FILE *fp;

/* We set these names rather indirectly for clarity in help() */

#define UNIXSTACKFILENAME ".cdpstack"
#define UNIXSCRIPTFILENAME ".cdpscript"
#define DOSSTACKFILENAME "C:/CDP.DAT"

#ifdef UNIX
char *home;
#define STACKFILENAME UNIXSTACKFILENAME
char *STACKFILE;
#define SCRIPTFILENAME UNIXSCRIPTFILENAME
char *SCRIPTFILE;
#else
#define STACKFILE DOSSTACKFILENAME
#endif

typedef struct Directories
   {
   char dir[MAXPATHLENGTH];
   struct Directories *next;
   } Directories;

int debugging = FALSE;

char destination[MAXPATHLENGTH] = { 0 };
char fullpathname[MAXPATHLENGTH] = { 0 };
char dir_to[MAXPATHLENGTH] = { 0 };
char cwdoriginal[MAXPATHLENGTH] = { 0 };
char drvoriginal[3] = { 0 };

int pushing = FALSE;

#define LETTERS 26

char options[LETTERS+3]; /* Ques and D and all locase alphas and a null */

char *optionmap[LETTERS] = /* Alphas only, and some are reserved. See help(). */
   {
#ifdef PATTERSON /* my default options */
#ifdef UNIX
   /* a */ NULL,
   /* b */ NULL,
   /* c */ NULL,
   /* d */ "~/diversa",
   /* e */ NULL,
   /* f */ NULL,
   /* g */ NULL,
   /* h */ /* Reserved, "help" */ NULL,
   /* i */ "~/incoming",
   /* j */ NULL,
   /* k */ "~/kit",
   /* l */ NULL,
   /* m */ "~/misc",
   /* n */ NULL,
   /* o */ NULL,
   /* p */ /* Reserved, "push" */ NULL,
   /* q */ NULL,
   /* r */ NULL,
   /* s */ "~/src",
   /* t */ "~/tmp",
   /* u */ NULL,
   /* v */ NULL,
   /* w */ NULL,
   /* x */ NULL,
   /* y */ NULL,
   /* z */ NULL
#else
   /* a */ NULL,
   /* b */ NULL,
   /* c */ "c:/william/comp",
   /* d */ "c:/william/diversa",
   /* e */ NULL,
   /* f */ NULL,
   /* g */ NULL,
   /* h */ /* Reserved, "help" */ NULL,
   /* i */ "c:/william/incoming",
   /* j */ NULL,
   /* k */ NULL,
   /* l */ "c:/william/language",
   /* m */ "c:/william/misc",
   /* n */ NULL,
   /* o */ NULL,
   /* p */ /* Reserved, "push" */ NULL,
   /* q */ "c:/william/iq",
   /* r */ NULL,
   /* s */ "c:/william/sources",
   /* t */ "/temp",
   /* u */ NULL,
   /* v */ NULL,
   /* w */ "c:/william",
   /* x */ NULL,
   /* y */ NULL,
   /* z */ NULL
#endif
#else /* your default options, which can be tailored to your liking. */
   /* a */ NULL,
   /* b */ NULL,
   /* c */ NULL,
   /* d */ "c:/dos",
   /* e */ NULL,
   /* f */ NULL,
   /* g */ NULL,
   /* h */ /* Reserved, "help" */ NULL,
   /* i */ "c:/incoming",
   /* j */ NULL,
   /* k */ NULL,
   /* l */ NULL,
   /* m */ "c:/misc",
   /* n */ NULL,
   /* o */ NULL,
   /* p */ /* Reserved, "push" */ NULL,
   /* q */ NULL,
   /* r */ NULL,
   /* s */ NULL,
   /* t */ "c:/temp",
   /* u */ NULL,
   /* v */ NULL,
   /* w */ NULL,
   /* x */ NULL,
   /* y */ NULL,
   /* z */ NULL
#endif
   };

/*****************************************************************************/

#define GETOPTERR(s,c) \
   if (opterr) \
      { \
      char errbuf[2]; \
      errbuf[0] = c; \
      errbuf[1] = '\n'; \
      (void) write(2,argv[0],(unsigned)strlen(argv[0])); \
      (void) write(2,s,(unsigned)strlen(s)); \
      (void) write(2,errbuf,2); \
      }

int opterr = 1;
int optind = 1;
int optopt;
char *optarg;

int mygetopt(argc,argv,opts)
   int argc;
   char **argv;
   char *opts;

   {

   static int sp = 1;
   register int c;
   register char *cp;

   if ( sp == 1 )
      {
      if ( optind>=argc || argv[optind][0]!='-' || argv[optind][1]=='\0' )
         {
         return(EOF);
         }
      else
         {
  if ( strcmp(argv[optind],"--") == (int)NULL ) /* 002 */
            {
            ++optind;
            return(EOF);
            }
         }
      }

   optopt = c = argv[optind][sp];

   if ( c==':' || (cp=strchr(opts,c))==NULL )
      {
      GETOPTERR(": illegal option -- ", c);
      if ( argv[optind][++sp] == '\0' )
         {
         ++optind;
         sp = 1;
         }
      return('?');
      }

   if ( *++cp == ':' )
      {
      if ( argv[optind][sp+1] != '\0' )
         {
         optarg = &argv[optind++][sp+1];
         }
      else
         {
         if ( ++optind >= argc )
            {
            GETOPTERR(": option requires an argument -- ",c);
            sp = 1;
            return('?');
            }
         else
            {
            optarg = argv[optind++];
            }
         }
      sp = 1;
      }
   else
      {
      if ( argv[optind][++sp] == '\0' )
         {
         sp = 1;
         ++optind;
         }
      optarg = NULL;
      }

   return(c);

   }

/*****************************************************************************/

#ifdef UNIX

int mkscript(mypath)
   char *mypath;
   {
   DEBUGMESSAGE(("mkscript...\n"));
   if ( NULL == ( fp = fopen(SCRIPTFILE, "w") ) )
      {
      printf("Unable to open %s\n", SCRIPTFILE);
      return 0;
      }
   fprintf(fp, "cd %s\n", mypath);
   fclose(fp);
   return 1;
   }

#endif

/*****************************************************************************/

/* Add the dir to the stack file */

void add_dir(dir_to_add)
   char *dir_to_add;
   {
#ifdef HI_TECH_C
   unsigned written;
#else
   int written;
#endif
   DEBUGMESSAGE(("add_dir... %s...\n",dir_to_add));
   DEBUGMESSAGE(("opening stackfile %s...\n",STACKFILE));
#ifdef UNIX
   if ( NULL == ( fp = fopen(STACKFILE,"a") ) )
#else
   if (!(fp = fopen(STACKFILE,"a")))
#endif
      {
      perror(STACKFILE);
      return;
      }
   written = fprintf(fp,"%s\n",dir_to_add);
   fclose(fp);
   return;
   }

/*****************************************************************************/

int read_dirs(Dirs)
   Directories *Dirs;
   {
   Directories *Top;
   DEBUGMESSAGE(("read_dirs...\n"));
   Top = Dirs;
   fp = fopen(STACKFILE, "r");
   if(!fp)
      {
      printf("directory stack is empty");
      UNIXCRLF;
      return 0;
      }
   while(!feof(fp))
      {
      if ( EOF != fscanf(fp, "%s", Dirs->dir) )
         {
         Dirs->next = (Directories *)malloc(sizeof(Directories));
         Dirs = Dirs->next;
         Dirs->next = NULL;
         }
      }
   Dirs = Top;
   fclose(fp);
   return 1;
   }

/*****************************************************************************/

int change_dir(Dir)
   Directories *Dir;
   {
   Directories *Top;
   DEBUGMESSAGE(("change_dir...\n"));
   Top = Dir;
   while(Dir->next != NULL)
      {
      strcpy(dir_to, Dir->dir);
      Dir = Dir->next;
      }
#ifdef UNIX
   mkscript(dir_to);
#endif
#ifdef _MSC_VER
   DEBUGMESSAGE(("switching to %s...\n",dir_to));
   _chdrive(toupper(dir_to[0])-'@');
   chdir(dir_to);
#endif
#ifdef HI_TECH_C
   DEBUGMESSAGE(("switching to %s...\n",dir_to));
   chdrv(dir_to);
   chdir(dir_to);
#endif
#ifdef UNKNOWN
   puts("I don't know how to chdrv and chdir.");
#endif
   Dir = Top;
   return 1;
   }

/*****************************************************************************/

int write_dirs(Dir)
   Directories *Dir;
   {
   Directories *Temp;
   DEBUGMESSAGE(("write_dirs...\n"));
   remove(STACKFILE);
   fp = fopen(STACKFILE, "w");
   if(!fp)
      {
      printf("Unable to open %s\n", STACKFILE);
      return 0;
      }
   while(Dir->next->next != NULL)
      {
      fprintf(fp, "%s\n", Dir->dir);
      Temp = Dir;
      Dir = Dir->next;
      free(Temp);
      }
   fclose(fp);
   return 1;
   }

/*****************************************************************************/

int check_fsize()
   {
   long length;
   DEBUGMESSAGE(("check_fsize...\n"));
   fp = fopen(STACKFILE, "r");
   fseek(fp, 0L, SEEK_END);
   length = ftell(fp);
   fclose(fp);
   if(length == 0)
      {
      DEBUGMESSAGE(("%s is empty; removing it...\n",STACKFILE));
      remove(STACKFILE);
      }
   return 1;
   }

/*****************************************************************************/

#ifdef UNIX

void STACKFILEliberation()
   {
   free(STACKFILE);
   }

#endif

/*****************************************************************************/

#ifdef UNIX

void SCRIPTFILEliberation()
   {
   free(SCRIPTFILE);
   }

#endif

/*****************************************************************************/

void configuration()
   {

   char varname[5];
   int x;
   char *p;
   int i;

#ifdef UNIX
   if ( NULL == ( home = getenv("HOME") ) )
      {
      puts("configuration: HOME not found in environment.");
      exit(1);
      }
   else
      {
      if ( NULL == ( STACKFILE = malloc(strlen(home)+1+strlen(STACKFILENAME)+1) ) )
         {
         perror("configuration: malloc");
         exit(1);
         }
      else
         {
         if (atexit(STACKFILEliberation))
            {
            puts("warning: atexit 1 failed");
            }
         strcpy(STACKFILE,home);
         strcat(STACKFILE,"/");
         strcat(STACKFILE,STACKFILENAME);
         }
      if ( NULL == ( SCRIPTFILE = malloc(strlen(home)+1+strlen(SCRIPTFILENAME)+1) ) )
         {
         perror("configuration: malloc");
         exit(1);
         }
      else
         {
         if (atexit(SCRIPTFILEliberation))
            {
            puts("warning: atexit 2 failed");
            }
         strcpy(SCRIPTFILE,home);
         strcat(SCRIPTFILE,"/");
         strcat(SCRIPTFILE,SCRIPTFILENAME);
         }
      }
#endif

   memset(options,0,sizeof(options));
   options[0] = '?';
   options[1] = 'D';
   i = 1;

   for ( strcpy(varname,"CDP?"), x = 0; x < LETTERS; x++ )
      {
      switch('a'+x)
         {
         case 'h':
         case 'p':
            options[++i] = 'a'+x;
            break;
         default:
            varname[3] = toupper(('a'+x));
            if ( NULL != ( p = getenv(varname) ) ) optionmap[x] = p;
            if ( NULL != optionmap[x] ) options[++i] = 'a'+x;
            break;
         }
      }
   }

/*****************************************************************************/

void fixslashes(p) /* convert backslashes to slashes */
   char *p;
   {
   for ( ; *p; ++p ) if ( '\\' == *p ) *p = '/';
   }

/*****************************************************************************/

void help(int verbose)
   {
   int x;
   if (verbose)
      {
      printf("\n %s\n",IDENTIFICATION);
      puts("\n Homepage: http://www.komputado.com/softvaro/misc.htm");
      puts("\n Usage: cdp [options] [drive:][path]");
      puts("\n This program can be used just like the familiar cd command... mostly...");
      puts(" except the DOS version doesn't print the current working directory if no");
      puts(" arguments are supplied, and the UNIX version doesn't take you home if no");
      puts(" arguments are supplied. Other than that... \"cdp \\temp\" takes you to \\temp");
      puts(" and \"cdp /tmp\" will take you to /tmp, as you'd expect.");
      puts("\n But, before the move is performed, cdp can stash your current location");
      puts(" in a safe place... \"cdp -p \\temp\" pushes the current info onto a stack");
      puts(" before taking you to the \\temp directory. Then you can cd and cdp all");
      puts(" over the place, and even \"cdp -p\" to other places. And when you're ready");
      puts(" to return to the last place you stashed, enter \"cdp\" alone to pop the most");
      puts(" recently pushed directory and move back to there. And \"cdp\" again to move");
      puts(" back to the one you pushed before that, and so on and so on.");
      puts("\n But cdp also offers another handy feature. Except for -p and -h, cdp");
      puts(" option letters can be defined for other oft-used directories. This can be");
      puts(" done in two ways. If you have a compiler handy, you can code them in and");
      puts(" recompile. Or you can define environment variables to define or replace");
      puts(" the ones which have already been compiled in. For example, if you do this");
      puts(" in DOS: \"set cdpt=\\temp\"; or this in UNIX: \"export CDPT=/tmp\"; then");
      puts(" \"cdp -t\" will always move you to that temporary directory (cdp doesn't");
      puts(" care whether you use slashes or backslashes). And in DOS you can also");
      puts(" specify a destination drive with cdp: \"cdp c:/backups\". Or just a drive");
      puts(" alone: \"cdp a:\" or \"cdp -p a:\".");
      puts("\n But, Wait, as they say, There's More! You can combine a relative path");
      puts(" designation (or several, though I rarely find that useful) with an option");
      puts(" For example, if cdp is set up so that \"-t\" moves you to \\temp, and you've");
      puts(" got a \"stuff\" directory in there, \"cdp -t stuff\" will move you over to");
      puts(" \\temp\\stuff. So, whenever more than one path argument is specified, it is");
      puts(" treated as relative to what's already been specified (and therefore must");
      puts(" not have a leading slash or a drive).");
      puts("\n The directory stack is kept in a temporary file...");
      printf("    DOS:  %s\n",DOSSTACKFILENAME);
      printf("    UNIX: $HOME/%s",UNIXSTACKFILENAME);
#ifdef UNIX
      printf(", currently %s\n",STACKFILE);
#else
      puts("");
#endif
      puts(" which is removed whenever it gets empty. (You can purge the stack at any time");
      puts(" by simply deleting this file.)");
      puts("\n UNIX Note:\n");
      puts("    Whereas the DOS version actually does the directory change, with UNIX");
      puts("    we must take a different approach; cdp instead creates a script...");
      printf("       $HOME/%s",UNIXSCRIPTFILENAME);
#ifdef UNIX
      printf(", currently %s\n",SCRIPTFILE);
#else
      puts("");
#endif
      puts("    which must then be sourced by the shell. For example, in .kshrc I have...");
      puts("\n        function cdp");
      puts("           {");
      puts("           if ~/bin/cdp $*");
      puts("           then");
      puts("              . ~/.cdpscript");
      puts("           fi");
      puts("           }");
      puts("\n    POSIX shell prefers something like this (to handle the case of");
      puts("    no argument, since it just bombs if you reference an environment");
      puts("    variable that hasn't been set)...");
      puts("\n       function cdp");
      puts("          {");
      puts("          if ~/bin/cdp ${*:-""}");
      puts("          then");
      puts("             . ~/.cdpscript");
      puts("          fi");
      puts("          }");
      puts("\n    If the script is created successfully, cdp returns 0 and the script is");
      puts("    run. If a valid script is not created for any reason, then cdp returns 1");
      puts("    (not very informative, but we just want to know whether to run the script).");
      puts("\n Examples:");
      puts("    cdp /incoming");
      puts("    cdp -p \\incoming");
      puts("    cdp -p a:/incoming");
      puts("    cdp -pt");
      puts("    cdp -i comp\\compilers");
      puts("    cdp -pi comp compilers");
      puts("    cdp");
      }
   printf("\n%s", verbose ? "Standard Options\n" : "" );
   puts("     ?   Short help");
   puts("    -?   Long help");
   puts("    -h   Long help");
   printf("    -D   Enable debug messaging (if compiled with debug messaging (it was");
#ifdef DEBUGGING
   puts("))");
#else
   puts("n't))");
#endif
   puts("    -p   Push current path to stack");
   if (verbose)
      {
      puts("\n As stated earlier, other alpha options can be re/defined by environment");
      puts(" variables, with names of the form CDPa, with 'a' being the option to be");
      puts(" altered. Current definitions, as compiled today and with whatever");
      puts(" environment variables set, are...");
      }
   for ( x=0; x<LETTERS; x++ ) if ( NULL != optionmap[x] ) printf("    -%c   Go to %s\n",'a'+x,optionmap[x]);
   UNIXCRLF;
   exit(1);
   }

/*****************************************************************************/

main(argc,argv)
   int argc;
   char **argv;

   {

   Directories *Dirs;

   int opt;

#ifdef UNKNOWN
   printf("This program was improperly compiled. Execution ends.\n");
   exit(1);
#endif

   configuration();

   while ( optind < argc )
      {
      while ( ( opt = mygetopt(argc,argv,options) ) != EOF )
         {
         switch(opt)
            {
            case 'D':
               debugging = TRUE;
               DEBUGNOTICE;
               break;
            case '?':
            case 'H':
            case 'h':
               help(1);
               break;
            case 'p':
               pushing = TRUE;
               break;
            default:
               if (destination[0])
                  {
                  printf("too many paths specified!");
                  UNIXCRLF;
                  exit(1);
                  }
               else
                  {
                  DEBUGMESSAGE(("setting destination to %s...\n",optionmap[tolower(opt)-'a']));
                  strcpy(destination,optionmap[tolower(opt)-'a']);
                  fixslashes(destination);
                  }
               break;
            }
         }
      if ( optind < argc )
         {
         switch(*argv[optind])
            {
            case '?':
               help(0);
               break;
            default:
               if ( ( strlen(destination) + strlen(argv[optind]) ) > ( MAXPATHLENGTH - 2 ) )
                  {
                  printf("path can be no longer than %d characters",MAXPATHLENGTH-1);
                  UNIXCRLF;
                  exit(1);
                  }
               else
                  {
                  fixslashes(argv[optind]);
                  if (destination[0])
                     {
                     if ( '/' == *argv[optind] )
                        {
                        printf("%s is not a relative path!",argv[optind]);
                        UNIXCRLF;
                        exit(1);
                        }
                     if ( '/' != destination[strlen(destination)-1] ) strcat(destination,"/");
                     }
                  DEBUGMESSAGE(("appending %s to destination...\n",argv[optind]));
                  strcat(destination,argv[optind]);
                  DEBUGMESSAGE(("destination is %s...\n",destination));
                  }
               break;
            }
         ++optind;
         }
      }

   if (destination[0])
      {
      DEBUGMESSAGE(("going...\n"));
      DEBUGMESSAGE(("destination is %s...\n",destination));
#ifdef _MSC_VER
      if ( NULL == getcwd(fullpathname,sizeof(fullpathname)-1) )
         {
         printf("getcwd failed");
         exit(1);
         }
      memset(drvoriginal,0,sizeof(drvoriginal));
      memcpy(drvoriginal,fullpathname,2);
      drvoriginal[0] = toupper(drvoriginal[0]);
      strcpy(cwdoriginal,fullpathname+2);
#endif
#ifdef HI_TECH_C
      strcpy(drvoriginal,getdrv());
      strcpy(cwdoriginal,getcwd(""));
      strcpy(fullpathname,drvoriginal);
      strcat(fullpathname,cwdoriginal);
#endif
#ifdef UNIX
      strcpy(fullpathname,getcwd(cwdoriginal,sizeof(cwdoriginal)));
#endif
#ifndef UNIX
      DEBUGMESSAGE(("current drive is %s...\n",drvoriginal));
      DEBUGMESSAGE(("current path is %s...\n",cwdoriginal));
#endif
      DEBUGMESSAGE(("current location is %s...\n",fullpathname));
      DEBUGMESSAGE(("switching to %s...\n",destination));
#ifndef UNIX
      if ( ':' == destination[1] )
         {
         if (!destination[2])
            {
            DEBUGMESSAGE(("appending trailing slash to drive specification...\n"));
            destination[2] = '/';
            }
#ifdef _MSC_VER
         if ( -1 == _chdrive(toupper(destination[0])-'@') )
#else
         if ( -1 == chdrv(destination) )
#endif
            {
            perror("chdrv");
            exit(1);
            }
         }
#endif
#ifdef UNIX
      mkscript(destination);
#else
      if ( -1 == chdir(destination) )
         {
#ifdef _MSC_VER
         _chdrive(toupper(drvoriginal[0])-'@');
#else
         chdrv(drvoriginal);
#endif
         perror("chdir");
         exit(1);
         }
#endif
      DEBUGMESSAGE(("stackfile is %s\n",STACKFILE));
      DEBUGMESSAGE(("fullpathname is %s\n",fullpathname));
      if (pushing) add_dir(fullpathname);
      }
   else
      {
      DEBUGMESSAGE(("coming...\n"));
      Dirs = (Directories *)malloc(sizeof(Directories));
      if(Dirs == NULL)
         {
         perror("malloc");
         exit(1);
         }
      if(!read_dirs(Dirs)) exit(1);
      change_dir(Dirs);
      write_dirs(Dirs);
      check_fsize();
      }

   exit(0);

   }
